#include "../eSkel.c"

/**************************/
/*     Pipeline Demo      */
/**************************/

/* Pipelined Gaussian elimination with local spread */

#define PROCS  8
#define STAGES 8
#define N 8

float wholea[N*(N+1)];

float mya[N+1]; 

int myrealrank;

float myx;

/* sequential read in and scattering of array */
void setuparray(FILE *infile)
{
  int i=0;

  if (myrank() == 0) {
    while (fscanf(infile, "%f", &wholea[i++])&&(i<N*(N+1)));
    if (i!=N*(N+1)) fprintf(stderr,"Problem reading input\n");
  }
  MPI_Scatter(wholea, N+1, MPI_FLOAT, mya, N+1, MPI_FLOAT, 0, mycomm());
}

void show (float *a, int n)
{
  int i, p;
  for (p=0; p<mycommsize(); p++) {
    if (p==myrank()) {
      printf("P%d has ", myrank());
      for (i=0; i<n; i++) {
        printf("%.2f ", a[i]);
      }
      printf("\n\n");
    }
    fflush(stdout);
    MPI_Barrier(mycomm());
  }
}

void reduce (void)
{
  static int startcol = 0;
  float *rowin;
  int ncols=N+1;
  int i;

  /* fprintf(stderr,"P%d starting stageproc\n",myrealrank); */

  while (rowin = Take(&ncols)) {
    /* pass on incoming update then work with it */
    if (myrealrank<N-1) Give(rowin, ncols);
    for (i=N; i>=startcol; i--) {
      mya[i] -= mya[startcol]*rowin[i];
    }
    startcol++;
  }
  for (i=startcol+1; i<N+1; i++) {
    mya[i] /= mya[startcol];
  }
  mya[startcol] = 1;
  if (myrealrank<N-1) Give(mya, N+1);
}

void backsub (void)
{
  static int startcol = N-1;
  float *knownx;
  int ncols=1;
  int i;
  
  while (knownx = Take(&ncols)) {
    /* pass on incoming update then work with it */
    if (myrealrank>0) Give(knownx, 1);
    mya[N] -= mya[startcol]* *knownx;
    startcol--;
  }
  /* I now know my variable's value */
  myx  = mya[N];
  if (myrealrank>0) Give(&myx, 1);
}


int main (int argc, char *argv[])
{
  spread_t spreads[STAGES+1];
  MPI_Datatype types[STAGES+1];
  void (*stages[STAGES])(void);
  int outmul;
  int i;

  FILE *infile;

  MPI_Init(&argc, &argv);
  SkelLibInit();
  myrealrank = myrank();
  if (myrank()==0) 
    if (!(infile = fopen(argv[1],"r"))) printf("Couldn't open %s\n", argv[1]);
  setuparray(infile);

  show(mya,N+1);


  for (i=0; i<STAGES; i++) {
    stages[i] = (void (*) (void)) reduce;
  }

  for (i=0; i<STAGES+1; i++) {
    spreads[i] = SPGLOBAL;
    types[i]   = MPI_FLOAT;
  }

  Pipeline (STAGES, stages, myrank(), spreads, types,
            NULL, 0, 0, NULL, 0, &outmul, 0, mycomm());

  for (i=0; i<STAGES; i++) {
    stages[i] = (void (*) (void)) backsub;
  }

  Pipeline (STAGES,  stages, N-myrank()-1, spreads, types,
            NULL, 0, 0, NULL, 0, &outmul, 0, mycomm());

  
  show(mya, N+1);
  
  show(&myx, 1);

  MPI_Finalize();
  return 0;
}
