#include "../eSkel.c"

/* Testing local spread pipeline */

#define PROCS  8
#define STAGES 3
#define POINTS 2
#define MULT 3

eSkel_atom_t *stageproc(eSkel_atom_t *datum)
{
  int i;
  int state = 1;
  int *data = (int *)(datum->data);
  int np = datum->len;

  if (myrank() == 0) { 
    for (i=0; i<np; i++) {
        data[i]+=state;
    }
    return datum;
  } else {
    return NULL; /* I'm not a stage root so do nothing for now */
  }
}


int main (int argc, char *argv[])
{
  int *points;
  int *results;
  int i, j, p, next;
  spread_t spreads[STAGES+1];
  MPI_Datatype types[STAGES+1];
  int outmul;
  int mystagenum;

  int mymult;

  eSkel_atom_t *(*stages[3])(eSkel_atom_t *);

  MPI_Init(&argc, &argv);
  SkelLibInit();

  points  = (int *) malloc (sizeof(int)*POINTS*MULT);
  results = (int *) malloc (sizeof(int)*POINTS*MULT);
  for (i=0; i<POINTS*MULT; i++) {
    points[i]  = myrank();
  }
  for (i=0; i<POINTS*MULT; i++) {
        results[i] = 99;
  }

  /* Set up all the paraphenalia */
  for (i=0; i<STAGES; i++) {
    stages[i] = stageproc;
  }

  for (i=0; i<STAGES+1; i++) {
    spreads[i] = SPLOCAL;
    types[i]   = MPI_INT;
  }

  if (myrank() < 3) 
    mystagenum = 0;
  else if (myrank() < 5) 
    mystagenum = 1;
  else
    mystagenum = 2;

  Pipeline1for1 (STAGES,  stages, mystagenum, spreads, types, 
                 (void *) points, POINTS, MULT, (void *) results, POINTS, 
                 &outmul, POINTS*MULT, mycomm());

  for (p=0; p<mycommsize(); p++) {
    MPI_Barrier (mycomm());
    if (myrank() == p) {
      printf("P%d has ",myrank());
      printf("points: ");

      next = 0;
      for (i=0; i<MULT; i++) {
        for (j=0; j<POINTS; j++) {
          printf("%d ", points[next++]);
        }
        printf("    ");
      }
      printf("\n");
      printf("      results: ");
      next = 0;
      for (i=0; i<outmul; i++) {
        for (j=0; j<POINTS; j++) {
          printf("%d ", results[next++]);
        }
        printf("    ");
      }
      printf("\n");
    }
  }    
  MPI_Finalize();
  return 0;
}
