#include "../eSkel.c"

#define P 4  // number of processes
#define S 2  // number of stages

eSkel_atom_t *timesten (eSkel_atom_t *datum) {

  int i;
  int *data = (int *)(datum->data);

  for (i=0; i<datum->len; i++) {
    data[i] *= 10;
  }
  return datum;

}

eSkel_atom_t *swap (eSkel_atom_t *datum) {

  MPI_Status status;

  int *data = (int *)(datum->data);

  MPI_Sendrecv_replace (data, datum->len, MPI_INT, (myrank()+1)%2, 0, 
                        (myrank()+1)%2, 0, mycomm(), &status);
  return datum;

}

int main (int argc, char *argv[]) {
  int *inputs;
  int *results;
  int *gathereddata = NULL;
  int i, inmul, outmul, inlen, outlen;
  spread_t spreads[S+1];
  MPI_Datatype types[S+1];
  eSkel_atom_t *(*stages[S])(eSkel_atom_t *);

  MPI_Init(&argc, &argv);
  SkelLibInit();

  inmul = 3;
  if (myrank() == 0) { //P0 creates all the data
    inputs = (int *) malloc (sizeof(int) * 12);
    for (i=0; i<12; i++) {
      inputs[i] = i+1;
    }
    inlen = 4;
  } else {
    inputs = NULL; inlen = 0;
  }
  results = (int *) malloc (sizeof(int)*3);
  outlen = 1;

  stages[0] = timesten; stages[1] = swap; 

  for (i=0; i<S+1; i++) {
    spreads[i] = SPGLOBAL;
    types[i]   = MPI_INT;
  }

  if (myrank() == 0) {
    printf("Inputs on P0: ");
    for (i=0; i<12; i++) {
      printf ("%d ", inputs[i]);
    }
    printf("\n");
    fflush(stdout);
  }

  Pipeline1for1 (S, stages, myrank()/S, spreads, types, (void *) inputs, inlen, inmul,
                 (void *) results, outlen, &outmul, 3, mycomm());

 
  if (myrank() == 0) {
    gathereddata = (int *) malloc (sizeof(int)*12);
  }

  for (i=0; i<outmul; i++) {
    MPI_Gather (&results[i], 1, MPI_INT, &gathereddata[i*4], 1, MPI_INT, 0, MPI_COMM_WORLD);
  }

  if (myrank() == 0) {
    printf("\nGathered results on P0: ");
    for (i=0; i<12; i++) {
      printf ("%d ", gathereddata[i]);
    }
    printf("\n");
  }

  MPI_Finalize();
  return 0;
}
