/* Simple farmer, run as a self contained process, ie unthreaded */
void farmerproc (void) 
{
  int i, j;

  int shutdowncount = 0;     /* How many workers have I told to shutdown           */
                             /* Assume workers behave - could enforce this in Take */
  MPI_Status status;
  int wid;                   /* id of the requesting worker */
  int target;
  int junk;

  while (shutdowncount < FrStack[Top]->context.sf.parentcommsize-1) {
    MPI_Recv(NULL, 0, MPI_BYTE, MPI_ANY_SOURCE, FARMTAKETAG, FrStack[Top]->context.sf.tofarmercomm, &status);
    wid = status.MPI_SOURCE;
    if (FrStack[Top]->context.sf.inbnext < FrStack[Top]->context.sf.allinmul) {
      /* send the next item to the worker */
      MPI_Send (&(FrStack[Top]->context.sf.inblens[FrStack[Top]->context.sf.inbnext]), 1, 
                MPI_INT, wid, NOSHUTDOWNTAG, FrStack[Top]->context.sf.fromfarmercomm); 
      MPI_Send (FrStack[Top]->context.sf.inbuff[FrStack[Top]->context.sf.inbnext], 
                FrStack[Top]->context.sf.inblens[FrStack[Top]->context.sf.inbnext],
                FrStack[Top]->context.sf.inty, 
                wid, NOSHUTDOWNTAG, FrStack[Top]->context.sf.fromfarmercomm); 
      /* Now send a source item tag */
      MPI_Send (&(FrStack[Top]->context.sf.inbnext), 1, MPI_INT, wid, NOSHUTDOWNTAG, FrStack[Top]->context.sf.fromfarmercomm); 
      FrStack[Top]->context.sf.inbnext++;
    } else {
      /* send the SHUTDOWNTAG to the worker root only */
      MPI_Send (&junk, 1, MPI_INT, wid, SHUTDOWNTAG, FrStack[Top]->context.sf.fromfarmercomm); 
      shutdowncount++;
    }
  }
}


void SimpleFarmv (void worker(void), void *in,  int *inlen, int inmul, spread_t inspr, MPI_Datatype inty, 
                  void *out, int *outlen, int *outmul, spread_t outspr, MPI_Datatype outty, MPI_Comm comm)
{
  MPI_Status stat;
  char *sndptr;

  char **woutbuff;        /* on worker for gathering the results */
  int  woutmult;          /* just a convenient copy of final outbnext on the worker */
  int  *woutlens;
                          /* the tags are already on the worker, for either spread */     


  char **foutbuff;        /* on comm root only, for gathering all results before final scatter */
  int  foutmult;
  int  *foutlens;
  int  *fouttags;

  int onewmult;
  int i, j, k, next;

  char * rcvptr;
  int outbuffspaceleft;
  int outbufflensspaceleft;

  int *tempptr;
  int temp;

  Top++;
  FrStack[Top] = (skel_t *) malloc (sizeof(skel_t));
  FrStack[Top]->skel = SFskel;
  FrStack[Top]->context.sf.parentcomm = comm;
  MPI_Comm_size(comm, &FrStack[Top]->context.sf.parentcommsize);
  MPI_Comm_rank(comm, &FrStack[Top]->context.sf.parentcommrank);

  MPI_Comm_dup (comm, &FrStack[Top]->context.sf.tofarmercomm);
  MPI_Comm_dup (comm, &FrStack[Top]->context.sf.fromfarmercomm);

  FrStack[Top]->context.sf.inspr      = inspr;  
  FrStack[Top]->context.sf.inty       = inty;   
  MPI_Type_size(inty, &FrStack[Top]->context.sf.insz);

  FrStack[Top]->context.sf.outspr      = outspr;  
  FrStack[Top]->context.sf.outty       = outty;   
  MPI_Type_size(outty, &FrStack[Top]->context.sf.outsz);

  /* calculate the worker groups and note                                */
  /*  - all group members and their roots, all as upper ids in comm      */
  /*  - do this on all comm processes, even though in the current impl.  */
  /*    only comm0 will need to know                                     */

  /* split off comms for the worker groups (farmer's one is redundant, farmer is P0) */
  MPI_Comm_split(comm, FrStack[Top]->context.sf.parentcommrank, 0, &FrStack[Top]->mycomm);
  MPI_Comm_rank(FrStack[Top]->mycomm, &FrStack[Top]->myrank);
  MPI_Comm_size(FrStack[Top]->mycomm, &FrStack[Top]->mycommsize);

  /* gather all the input data into the inbuff on the farmer */
  if (FrStack[Top]->context.sf.inspr == SPGLOBAL) {
    if (FrStack[Top]->context.sf.parentcommrank == 0) {
      FrStack[Top]->context.sf.inblens = (int *) malloc (sizeof(int) * inmul);
      FrStack[Top]->context.sf.inbuff  = (char **) malloc (sizeof(char *)*inmul);
    }
    sndptr = (char *) in;
    for (i=0; i<inmul; i++) {
      if (FrStack[Top]->context.sf.parentcommrank == 0) {
        FrStack[Top]->context.sf.inblens[i] = -1;
        Glob_Unit_Gather (comm, FrStack[Top]->context.sf.parentcommsize, 
                          FrStack[Top]->context.sf.parentcommrank,
                          0, FrStack[Top]->context.sf.inty, FrStack[Top]->context.sf.insz,
                          sndptr, inlen[i], 
                          &(FrStack[Top]->context.sf.inbuff[i]), &FrStack[Top]->context.sf.inblens[i]);
      } else {
        Glob_Unit_Gather (comm, FrStack[Top]->context.sf.parentcommsize, 
                          FrStack[Top]->context.sf.parentcommrank,
                          0, FrStack[Top]->context.sf.inty, FrStack[Top]->context.sf.insz,
                          sndptr, inlen[i], NULL, NULL);
      }
      sndptr = sndptr + (FrStack[Top]->context.sf.insz * inlen[i]);
    }
    FrStack[Top]->context.sf.inbnext = 0;      /* set up "pointer" for Take */
    FrStack[Top]->context.sf.allinmul = inmul; 
  } else if (FrStack[Top]->context.sf.inspr == SPLOCAL) {
    Local_Mult_Gather (comm, FrStack[Top]->context.sf.parentcommsize, FrStack[Top]->context.sf.parentcommrank,
                       0, FrStack[Top]->context.sf.inty, FrStack[Top]->context.sf.insz,
                       in, inmul, inlen,
                       &FrStack[Top]->context.sf.inbuff, &FrStack[Top]->context.sf.allinmul,
                       &FrStack[Top]->context.sf.inblens);
    FrStack[Top]->context.sf.inbnext = 0; /* set up "pointer" for Take */
  }

  /*  set up the outbuffs in each worker */
  if  (FrStack[Top]->context.sf.parentcommrank != 0) {
      FrStack[Top]->context.sf.outblens = (int *)  malloc (sizeof(int)*MAXOUTMUL);
      FrStack[Top]->context.sf.outbtags = (int *)  malloc (sizeof(int)*MAXOUTMUL);
      FrStack[Top]->context.sf.outbuff  = (char **)malloc (sizeof(char *)*MAXOUTMUL);
  } else {
      FrStack[Top]->context.sf.outblens = NULL;
      FrStack[Top]->context.sf.outbtags = NULL;
      FrStack[Top]->context.sf.outbuff  = NULL;
  }
  FrStack[Top]->context.sf.outbcount = 0;  /* on the farmer too, just for the reduction later */

  /* call the farmer and the workers */
  if ( FrStack[Top]->context.sf.parentcommrank == 0) {
    farmerproc();
  } else {
    worker();
  }

  MPI_Barrier(comm);                      /* As with "farm", think about whether we really need this */

  /* gather/scatter the results from lower to upper level */
  /* this is why Given items need to be tagged with their Take item id*/
  
  woutmult = FrStack[Top]->context.sf.outbcount; /* copy for convenience only                 */
  woutbuff = FrStack[Top]->context.sf.outbuff;   /* already have all the data in outbuff      */
  woutlens = FrStack[Top]->context.sf.outblens;  /* so just point to it                       */

  if (FrStack[Top]->context.sf.parentcommrank == 0) { /* the farmer) */
    /* Count overall mult and set up data structure */
    MPI_Reduce (&woutmult, &foutmult, 1, MPI_INT, MPI_SUM, 0, comm);
    foutbuff = (char **) malloc (sizeof(char *) * foutmult);
    foutlens = (int *) malloc   (sizeof(int)    * foutmult);
    fouttags = (int *) malloc   (sizeof(int)    * foutmult);
    next=0;
    for (i=1; i<FrStack[Top]->context.sf.parentcommsize; i++) {
      MPI_Recv(&onewmult, 1, MPI_INT, i, MPI_ANY_TAG, comm, &stat);
      for (j=0; j<onewmult; j++) {
        MPI_Recv(&foutlens[next], 1, MPI_INT, i, MPI_ANY_TAG, comm, &stat);
        foutbuff[next] = (char *) malloc (FrStack[Top]->context.sf.outsz*foutlens[next]);
        MPI_Recv(foutbuff[next], foutlens[next], outty, i, MPI_ANY_TAG, comm, &stat); 
        MPI_Recv(&fouttags[next], 1, MPI_INT, i, MPI_ANY_TAG, comm, &stat);
        next++;
      }
    }
  } else {  /* a worker */
    MPI_Reduce (&woutmult, &foutmult, 1, MPI_INT, MPI_SUM, 0, comm);
    MPI_Send (&woutmult, 1, MPI_INT, 0, 0, comm); 
    for (j=0; j < woutmult; j++) {
      MPI_Send(&woutlens[j], 1, MPI_INT, 0, 0, comm);
      MPI_Send(woutbuff[j], woutlens[j], outty, 0, 0, comm); 
      MPI_Send(&FrStack[Top]->context.sf.outbtags[j], 1, MPI_INT, 0, 0, comm);
    }
  }

  /* farmer root sorts outputs by tag (stably wrt generated order per tag) */
  if (FrStack[Top]->context.sf.parentcommrank == 0) {
    stable_sort_by_tag(foutmult, fouttags, (int **)foutbuff, foutlens);
  }

  /* And scatter the results back at the upper level */
  if (outspr == SPGLOBAL) {
    rcvptr = (char *) out;
    outbuffspaceleft     = outlen[0];
    outbufflensspaceleft = outlen[1];
    /* broadcast actual multiplicity of output */
    MPI_Bcast(&foutmult, 1, MPI_INT, 0, comm);
    for (i=0; i< foutmult; i++) {
      temp = outbuffspaceleft;
      if (FrStack[Top]->context.sf.parentcommrank == 0) {
        Glob_Unit_Scatter (comm, FrStack[Top]->context.sf.parentcommsize, 
                           FrStack[Top]->context.sf.parentcommrank, 
                           0, outty, FrStack[Top]->context.sf.outsz,
                           (void *) foutbuff[i], foutlens[i], &rcvptr, &temp);
      } else {
        Glob_Unit_Scatter (comm, FrStack[Top]->context.sf.parentcommsize, 
                           FrStack[Top]->context.sf.parentcommrank, 
                           0, outty, FrStack[Top]->context.sf.outsz,
                           NULL, 0, &rcvptr, &temp);
      }
      outbuffspaceleft -= temp;
      if (i<outbufflensspaceleft) {
        outlen[i] = temp;
      } else {
       fprintf(stderr, "No more space for farm outbufflens\n");
       exit(0);
      }
      rcvptr += (temp*FrStack[Top]->context.sf.outsz);
    }
    *outmul = foutmult;
  } else {
    /* distribute local spread results */
    Local_Mult_Scatter (comm, FrStack[Top]->context.sf.parentcommsize, FrStack[Top]->context.sf.parentcommrank,
                        0, outty, FrStack[Top]->context.sf.outsz,
                        (void **) foutbuff, foutmult, 
                        foutlens, (char *) out, outmul, outlen);
  }

  Top--;
  return;
}

